%% Example:
% Compute an induced L2 gain from d to v for an uncertain system with a
% single (scalar) real uncertainty. This example is taken from Fetzer,
% Scherer, and Veenman, "Invariance with dynamic multipliers", submitted
% to TAC, 2017.

%% Model G
AG = [-2 -1 -1; 1 0 0.1; 0 1 0];
BG = [1 2; 0 -0.1; -0.1 0.2];
CG = [-1.1 0.5 0.1];
DG = [0 1];
nG = size(AG,1);

%% Multiplier Psi
v = 2;              % Order of psiv
kappa = 0.4;        % Bound on perturbation
%kappa = 0.8;

%kappa = 1.05;  % v=1, feas boundary is kappa = 1.07

% Create minimal realization for:
%    psiv(s) = [1 H H^2 ... H^v]^T
H = ss(-1,1,1,0);
psiv = ss(1);
for i1=1:v
    psiv = [eye(i1); zeros(1,i1-1) H]*psiv;
end

% Create filter Psi
Psi = blkdiag(kappa*psiv,psiv);
[APsi,BPsi,CPsi,DPsi] = ssdata(Psi);
nPsi = size(APsi,1);

%% Extended System
AA = [AG zeros(nG,nPsi); BPsi(:,1)*CG APsi];
BB1 = [BG(:,1); BPsi(:,1)*DG(1,1)+BPsi(:,2)];
BB2 = [BG(:,2); BPsi(:,1)*DG(1,2)];
CC = [DPsi(:,1)*CG CPsi];
DD1 = DPsi(:,1)*DG(1,1)+DPsi(:,2);
DD2 = DPsi(:,1)*DG(1,2);

%% Solve LMI for Induced L2 Gain
cvx_begin sdp quiet
    %cvx_solver sedumi;

    % Create variables
    variable P(nG+nPsi,nG+nPsi) symmetric;
    variable M11(v+1,v+1) symmetric;
    variable gsqinv;
    if v>0
        variable M12(v+1,v+1) skew_symmetric;
        variable Y22(nPsi,nPsi) symmetric;
    else
        M12 = 0;
        Y22 = [];
    end
    M = [M11 M12; M12' -M11];

    % Main Dissipation Inequality LMI
    [AA'*P+P*AA P*[BB1 BB2]; [BB1 BB2]'*P zeros(2)] ...
        + blkdiag(zeros(nG+nPsi+1),-1) ...
        + gsqinv*[CC DD1 DD2]'*[CC DD1 DD2] ...
        + [CC DD1 DD2]'*M*[CC DD1 DD2] <= 0;

    if v>0
        % LMI to ensure PI11>0 and PI22<0
        [APsi'*Y22+Y22*APsi Y22*BPsi(:,2); BPsi(:,2)'*Y22 0] + ...
            [CPsi DPsi(:,2)]'*M*[CPsi DPsi(:,2)] <= 0;
    end

    % Objective function
    maximize( gsqinv );
cvx_end

if isfinite(cvx_optval)
    % Valid solution. Should have cvx_optval = trace(H)
    gsq = 1/gsqinv;
    P = full(P);
    M = full(M);
    fprintf('\n Optimal Cost =  %4.3f',gsq);
else
    error('Infeasible Problem');
end
fprintf('\n\n');

%% Compute optimal IQC cost and the corresponding stabilizing solution.
Q = CPsi'*M*CPsi;
R = DPsi'*M*DPsi;
S = CPsi'*M*DPsi;
X = care(APsi,BPsi,Q,R,S);
K = inv(R)*(X*BPsi+S)';
