%% Example:
% Compute an invariant set for an uncertain system with a single (scalar)
% real uncertainty.  This example is taken from Fetzer, Scherer, and
% Veenman, "Invariance with dynamic multipliers", submitted to TAC, 2017.

%% Model G
AG = [-2 -1 -1; 1 0 0.1; 0 1 0];
BG = [1 2; 0 -0.1; -0.1 0.2];
CG = [-1.1 0.5 0.1];
DG = [0 1];
nG = size(AG,1);

%% Robust Stability Boundary via Mu analysis
% Note: ROBSTAB exists in R2016b and more recent versions
runStabMarg = false;
if runStabMarg
    del = ureal('del',0);
    G = ss(AG,BG,CG,DG);
    Gunc = lft(del,G);
    [stabmarg,WCU] = robstab(Gunc)
    
    pole( usubs(Gunc,WCU) )
end

%% Multiplier Psi
v = 2;              % Order of psiv
kappa = 0.4;        % Bound on perturbation
%kappa = 0.8;
%kappa = 1.05

% Create minimal realization for:
%    psiv(s) = [1 H H^2 ... H^v]^T
H = ss(-1,1,1,0);
psiv = ss(1);
for i1=1:v
    psiv = [eye(i1); zeros(1,i1-1) H]*psiv;
end

% Create filter Psi
Psi = blkdiag(kappa*psiv,psiv);
[APsi,BPsi,CPsi,DPsi] = ssdata(Psi);
nPsi = size(APsi,1);

%% Extended System
AA = [AG zeros(nG,nPsi); BPsi(:,1)*CG APsi];
BB1 = [BG(:,1); BPsi(:,1)*DG(1,1)+BPsi(:,2)];
BB2 = [BG(:,2); BPsi(:,1)*DG(1,2)];
CC = [DPsi(:,1)*CG CPsi];
DD1 = DPsi(:,1)*DG(1,1)+DPsi(:,2);
DD2 = DPsi(:,1)*DG(1,2);

%% Solve LMI in Iteration
Niter = 5;
OptCost = zeros(Niter,1);
K = zeros(2,nPsi);
for i=1:Niter
    % Solve LMI
    cvx_begin sdp quiet
        %cvx_solver sedumi;

        % Create variables
        variable P(nG+nPsi,nG+nPsi) symmetric;
        variable M11(v+1,v+1) symmetric;
        variable H(nG,nG) symmetric;
        if v>0
            variable M12(v+1,v+1) skew_symmetric;
            variable Y22(nPsi,nPsi) symmetric;
            variable Y22til(nPsi,nPsi) symmetric;
        else
            M12 = 0;
            Y22 = [];
            Y22til = [];
        end
        M = [M11 M12; M12' -M11];

        % Main Dissipation Inequality LMI
        [AA'*P+P*AA P*[BB1 BB2]; [BB1 BB2]'*P zeros(2)] ...
            + blkdiag(zeros(nG+nPsi+1),-1) ...
            + [CC DD1 DD2]'*M*[CC DD1 DD2] <= 0;

        if v>0
            % LMI to compute soft IQC bound
            [(APsi-BPsi*K)'*Y22til+Y22til*(APsi-BPsi*K) ...
                Y22til*BPsi(:,2); BPsi(:,2)'*Y22til 0] + ...
                [(CPsi-DPsi*K) DPsi(:,2)]'*M*[(CPsi-DPsi*K) DPsi(:,2)] <= 0;

            % LMI to ensure PI11>0 and PI22<0
            [APsi'*Y22+Y22*APsi Y22*BPsi(:,2); BPsi(:,2)'*Y22 0] + ...
                [CPsi DPsi(:,2)]'*M*[CPsi DPsi(:,2)] <= 0;
        end

        % Specify LMI to bound invariant set
        blkdiag(H,P) + [zeros(nG) eye(nG) zeros(nG,nPsi); ...
            eye(nG) zeros(nG) zeros(nG,nPsi); ...
            zeros(nPsi,nG) zeros(nPsi,nG) -Y22til] >= 0;

        % Objective function
        minimize( trace(H) );
    cvx_end
    
    if isfinite(cvx_optval)
        % Valid solution. Should have cvx_optval = trace(H)
        P = full(P);
        H = full(H);
        M = full(M);
        fprintf('\n i=%d  Optimal Cost =  %4.3f',i,trace(H));
    else
        error('Infeasible Problem');
    end
    OptCost(i) = trace(H);
    
    % Compute feedback gain for soft IQC bound
    if i<Niter
        % Compute optimal IQC cost from previous iteration and the
        % corresponding stabilizing solution.
        Q = CPsi'*M*CPsi;
        R = DPsi'*M*DPsi;
        S = CPsi'*M*DPsi;
        X = care(APsi,BPsi,Q,R,S);
        K = inv(R)*(X*BPsi+S)';
    end
end
fprintf('\n\n');

%% Verify Final Results
runVerifyResults = false;
if runVerifyResults
    % Form final PI and verify that it is valid for |del|<=kappa:
    % PI(1,1)>0, PI(2,2)<0, |PI(2,2)| = kappa^2*|PI(1,1)|
    % and PI(1,2)=-PI(1,2)'
    PI = Psi'*M*Psi;
    
    w = [0 logspace(-2,2,100) inf];
    PI11 = freqresp(PI(1,1),w);
    fprintf('\n-----Verify PI is valid multiplier for real delta\n');
    fprintf('min(real(PI11)) = %4.3f \t max(imag(PI11)) = %4.3f\n',...
        min(real(PI11(:))),max(imag(PI11(:))) );
    PI22 = freqresp(PI(2,2),w);
    fprintf('max(real(PI22)) = %4.3f \t max(imag(PI22)) = %4.3f\n',...
        max(real(PI22(:))),max(imag(PI22(:))) );
    fprintf('inf norm of PI(1,1)+kappa^2*PI(2,2) = %4.3f\n', ...
        norm( PI(1,1)+kappa^2*PI(2,2), inf) );
    
    fprintf('inf norm of PI(1,2)+PI(1,2)~ = %4.3f\n',...
        norm( PI(1,2)+PI(1,2)', inf))
    
    
    % Compute optimal IQC cost from final iteration and the
    % corresponding stabilizing solution.
    Q = CPsi'*M*CPsi;
    R = DPsi'*M*DPsi;
    S = CPsi'*M*DPsi;
    X = care(APsi,BPsi,Q,R,S);
    K = inv(R)*(X*BPsi+S)';
    
    % Verify that this IQC yields the final cost
    % Resolve the SDP with IQC (Psi,M) held fixed using ARE bound X
    fprintf('\n-----Resolve SDP with (Psi,M) held fixed\n');
    cvx_begin sdp quiet
        %cvx_solver sedumi;

        % Create variables
        variable P(nG+nPsi,nG+nPsi) symmetric;
        variable H(nG,nG) symmetric;

        % Main Dissipation Inequality LMI
        [AA'*P+P*AA P*[BB1 BB2]; [BB1 BB2]'*P zeros(2)] ...
            + blkdiag(zeros(nG+nPsi+1),-1) ...
            + [CC DD1 DD2]'*M*[CC DD1 DD2] <= 0;

        % Specify LMI to bound invariant set
        blkdiag(H,P) + [zeros(nG) eye(nG) zeros(nG,nPsi); ...
            eye(nG) zeros(nG) zeros(nG,nPsi); ...
            zeros(nPsi,nG) zeros(nPsi,nG) -X] >= 0;

        % Objective function
        minimize( trace(H) );
    cvx_end
    fprintf('(Verify #1)  Optimal Cost =  %4.3f\n',trace(H));
    
    % J-spectral factorization of multiplier
    fprintf('\n-----J spectral factorization\n');
    [U,T] = schur(R);
    ev = real( diag(T) );
    pidx = find( ev>0 );
    qidx = find( ev<0 );
    W = diag( sqrt([ev(pidx); -ev(qidx)]) );
    W = W*U(:,[pidx;qidx])';
    p = length(pidx);
    q = length(qidx);
    J = blkdiag( eye(p), -eye(q) );
    PsiJ = ss(APsi,BPsi,J*inv(W')*(BPsi'*X+S'),W);
    fprintf('inf norm of PI-PsiJ~ J PsiJ = %4.3f\n',...
        norm(PI - PsiJ'*J*PsiJ,inf));

    % Resolve SDP with J-spectral factorization
    fprintf('\n-----Resolve SDP with J-spectral Factorization\n');
    [APsi,BPsi,CPsi,DPsi] = ssdata(PsiJ);
    AA = [AG zeros(nG,nPsi); BPsi(:,1)*CG APsi];
    BB1 = [BG(:,1); BPsi(:,1)*DG(1,1)+BPsi(:,2)];
    BB2 = [BG(:,2); BPsi(:,1)*DG(1,2)];
    CC = [DPsi(:,1)*CG CPsi];
    DD1 = DPsi(:,1)*DG(1,1)+DPsi(:,2);
    DD2 = DPsi(:,1)*DG(1,2);
    
    cvx_begin sdp quiet
        %cvx_solver sedumi;

        % Create variables
        variable P(nG+nPsi,nG+nPsi) symmetric;
        variable H(nG,nG) symmetric;

        % Main Dissipation Inequality LMI
        [AA'*P+P*AA P*[BB1 BB2]; [BB1 BB2]'*P zeros(2)] ...
            + blkdiag(zeros(nG+nPsi+1),-1) ...
            + [CC DD1 DD2]'*J*[CC DD1 DD2] <= 0;

        % Specify LMI to bound invariant set
        blkdiag(H,P) + [zeros(nG) eye(nG) zeros(nG,nPsi); ...
            eye(nG) zeros(nG) zeros(nG,nPsi); ...
            zeros(nPsi,nG) zeros(nPsi,nG) zeros(nPsi)] >= 0;

        % Objective function
        minimize( trace(H) );
    cvx_end
    fprintf('(Verify #2)  Optimal Cost =  %4.3f\n',trace(H));    
end

